import threading

from abc import ABC, abstractmethod


class CoCoEntity(ABC):

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def online(self):
        return self._online

    @property
    def model(self):
        return self._model

    @property
    def type(self):
        return self._type

    @property
    def profile_creation_id(self):
        return self._profile_creation_id

    @property
    def on_change(self):
        return self._on_change

    @on_change.setter
    def on_change(self, func):
        with self._callback_mutex:
            self._on_change = func

    def __init__(self, dev, callback_container, client, profile_creation_id):
        self._client = client
        self._profile_creation_id = profile_creation_id
        self._uuid = dev['Uuid']
        self._name = None
        self._online = None
        self._model = None
        self._type = None
        self._state = None
        self._callback_mutex = threading.RLock()
        self._on_state_change = None
        self._callback_mutex = threading.RLock()
        self._on_change = None
        self._callback_container = None
        self._after_update_callback = None

    def update_dev(self, dev, callback_container=None):
        has_changed = False
        if 'Name' in dev and self._name != dev['Name']:
            self._name = dev['Name']
            has_changed = True
        if 'DisplayName' in dev and self._name != dev['DisplayName']:
            self._name = dev['DisplayName']
            has_changed = True
        if 'Online' in dev and self._online != (dev['Online'] == 'True'):
            self._online = dev['Online'] == 'True'
            has_changed = True
        if 'Model' in dev and self._model != dev['Model']:
            self._model = dev['Model']
            has_changed = True
        if 'Type' in dev and self._type != dev['Type']:
            self._type = dev['Type']
            has_changed = True
        if callback_container:
            self._callback_container = callback_container
            if 'callbackHolder' in self._callback_container:
                self._callback_container['callbackHolder'] = self._update
                has_changed = True
        return has_changed

    @abstractmethod
    def _update(self, dev):
        pass

    def _state_changed(self):
        self.on_change()




import json
import logging
import os
from typing import List, Callable

import paho.mqtt.client as mqtt

from .helpers import extract_devices
from .const import MQTT_PROTOCOL, MQTT_TRANSPORT, MQTT_TOPIC_PUBLIC_RSP, MQTT_TOPIC_SUFFIX_RSP, \
    MQTT_TOPIC_SUFFIX_SYS_EVT, MQTT_TOPIC_SUFFIX_CMD, MQTT_TOPIC_SUFFIX_EVT, MQTT_TOPIC_PUBLIC_CMD, MQTT_RC_CODES
from .coco_light import CoCoLight
from .coco_switch import CoCoSwitch

_LOGGER = logging.getLogger(__name__)


class CoCo:
    def __init__(self, address, username, password, port=8883, ca_path=None, switches_as_lights=False):
        if ca_path is None:
            ca_path = os.path.dirname(os.path.realpath(__file__)) + '/coco_ca.pem'
        client = mqtt.Client(protocol=MQTT_PROTOCOL, transport=MQTT_TRANSPORT)
        client.username_pw_set(username, password)
        client.tls_set(ca_path)
        client.tls_insecure_set(True)
        self._client = client
        self._address = address
        self._port = port
        self._profile_creation_id = username
        self._all_devices = None
        self._device_callbacks = {}
        self._switches_as_lights = switches_as_lights
        self._lights = None
        self._lights_callback = Callable[[List[CoCoLight]], None]
        self._switches = None
        self._switches_callback = Callable[[List[CoCoSwitch]], None]
        self._system_info = None
        self._system_info_callback = Callable[[List[CoCoSwitch]], None]

    def __del__(self):
        self._client.disconnect()

    def connect(self):

        def _on_message(client, userdata, message):
            topic = message.topic
            response = json.loads(message.payload)
            if topic == MQTT_TOPIC_PUBLIC_RSP and response['Method'] == 'systeminfo.publish':
                self._system_info = response
                self._system_info_callback(self._system_info)
                return
            if topic == (self._profile_creation_id + MQTT_TOPIC_SUFFIX_RSP):
                if response['Method'] == 'devices.list':
                    client.unsubscribe(self._profile_creation_id + MQTT_TOPIC_SUFFIX_RSP)
                    devices = extract_devices(response)
                    existing_uuids = list(self._device_callbacks.keys())

                    valid_switches = ['socket', 'switched-generic']
                    valid_lights = ['light', 'dimmer']
                    for x in devices:
                        if x['Uuid'] not in existing_uuids:
                            self._device_callbacks[x['Uuid']] = {'callbackHolder': None, 'entity': None}

                    lights = [x for x in devices if
                              (x['Model'] in valid_lights)
                              or (self._switches_as_lights
                                  and (x['Model'] in valid_switches))
                              ]
                    if not self._switches_as_lights:
                        switches = [x for x in devices if x['Model'] in valid_switches]
                    else:
                        switches = []

                    self._lights = []
                    for light in lights:
                        if self._device_callbacks[light['Uuid']] and self._device_callbacks[light['Uuid']]['entity'] and \
                                self._device_callbacks[light['Uuid']]['entity'].uuid:
                            self._device_callbacks[light['Uuid']]['entity'].update_dev(light)
                        else:
                            self._device_callbacks[light['Uuid']]['entity'] = CoCoLight(light, self._device_callbacks[
                                light['Uuid']], self._client, self._profile_creation_id)
                        self._lights.append(self._device_callbacks[light['Uuid']]['entity'])
                    self._switches = []
                    for switch in switches:
                        if self._device_callbacks[switch['Uuid']] and self._device_callbacks[switch['Uuid']][
                            'entity'] and \
                                self._device_callbacks[switch['Uuid']]['entity'].uuid:
                            self._device_callbacks[switch['Uuid']]['entity'].update_dev(switch)
                        else:
                            self._device_callbacks[switch['Uuid']]['entity'] = CoCoSwitch(switch,
                                                                                          self._device_callbacks[
                                                                                              switch['Uuid']],
                                                                                          self._client,
                                                                                          self._profile_creation_id)
                        self._switches.append(self._device_callbacks[switch['Uuid']]['entity'])

                    self._lights_callback(self._lights)
                    self._switches_callback(self._switches)
                return
            if topic == (self._profile_creation_id + MQTT_TOPIC_SUFFIX_SYS_EVT) \
                    and response['Method'] == 'systeminfo.published':
                # If the connected controller publishes sysinfo... we expect something to have changed.
                client.subscribe(self._profile_creation_id + MQTT_TOPIC_SUFFIX_RSP, qos=1)
                client.publish(self._profile_creation_id + MQTT_TOPIC_SUFFIX_CMD, '{"Method":"devices.list"}', 1)
                return
            if topic == (self._profile_creation_id + MQTT_TOPIC_SUFFIX_EVT) \
                    and (response['Method'] == 'devices.status' or response['Method'] == 'devices.changed'):
                devices = extract_devices(response)
                for device in devices:
                    if device['Uuid'] and self._device_callbacks[device['Uuid']]:
                        self._device_callbacks[device['Uuid']]['callbackHolder'](device)

                return

        def _on_connect(client, userdata, flags, rc):
            if rc == 0:
                _LOGGER.info('Connected!')
                client.subscribe(self._profile_creation_id + MQTT_TOPIC_SUFFIX_RSP, qos=1)
                client.subscribe(MQTT_TOPIC_PUBLIC_RSP, qos=1)
                client.subscribe(self._profile_creation_id + MQTT_TOPIC_SUFFIX_EVT, qos=1)
                client.subscribe(self._profile_creation_id + MQTT_TOPIC_SUFFIX_SYS_EVT, qos=1)
                client.publish(MQTT_TOPIC_PUBLIC_CMD, '{"Method":"systeminfo.publish"}', 1)
                client.publish(self._profile_creation_id + MQTT_TOPIC_SUFFIX_CMD, '{"Method":"devices.list"}', 1)
            elif MQTT_RC_CODES[rc]:
                raise Exception(MQTT_RC_CODES[rc])
            else:
                raise Exception('Unknown error')

        def _on_disconnect(client, userdata, rc):
            self._client._update_inflight()
            _LOGGER.warning('Disconnected')
            for uuid, device_callback in self._device_callbacks.items():
                offline = {'Online': 'False', 'Uuid': uuid}
                device_callback['callbackHolder'](offline)

        self._client.on_message = _on_message
        self._client.on_connect = _on_connect
        self._client.on_disconnect = _on_disconnect

        self._client.reconnect_delay_set(min_delay=3, max_delay=20)
        self._client.connect_async(self._address, self._port)
        self._client.loop_start()

    def disconnect(self):
        self._client.loop_stop()
        self._client.disconnect()

    def get_systeminfo(self, callback):
        self._system_info_callback = callback
        if self._system_info:
            self._system_info_callback(self._system_info)

    def get_lights(self, callback):
        self._lights_callback = callback
        if self._lights:
            self._lights_callback(self._lights)

    def get_switches(self, callback):
        self._switches_callback = callback
        if self._switches:
            self._switches_callback(self._switches)

    def _emit_switches(self):
        self._switches_callback(self._switches)



import binascii
import select
import threading
import socket
import netifaces
from getmac import get_mac_address


class CoCoDiscover:
    """CoCoDiscover will help you discover NHC2.
    It will also tell you about NHC1, but the result will differ.

    You create CoCoDiscover, passing along a callback an the time you max want to wait.
    By default we wait 3 seconds.
    
    For every result with matching header the callback is called,
    with the address, mac-address and a boolean if it's a NHC2.
    """

    def __init__(self, on_discover, on_done):
        self._get_broadcast_ips()

        self._thread = threading.Thread(target=self._scan_for_nhc)
        self._on_discover = on_discover
        self._on_done = on_done
        self._scan_time = 5
        self._thread.start()

    def _get_broadcast_ips(self):
        interfaces = netifaces.interfaces()
        return filter(lambda x: x,
                              map(lambda x: netifaces.ifaddresses(x).get(netifaces.AF_INET)[0].get('broadcast') if (
                                      (netifaces.AF_INET in netifaces.ifaddresses(x))
                                      and ('broadcast' in netifaces.ifaddresses(x).get(netifaces.AF_INET)[0])

                              ) else None, interfaces))

    def _scan_for_nhc(self):
        server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        """ We search for all broadcast ip4s, so that we don't only search the main interface """
        broadcast_ips = self._get_broadcast_ips()
        for broadcast_ip in broadcast_ips:
            server.sendto(bytes([0x44]), (broadcast_ip, 10000))
        server.setblocking(0)
        loops = 0

        while loops < 2:
            loops = loops + 1
            ready = select.select([server], [], [], self._scan_time)
            if ready[0]:
                data, addr = server.recvfrom(4096)
                if data[0] == 0x44:  # NHC2 Header
                    is_nhc2 = (len(data) >= 16) and (data[15] == 0x02)
                    mac = get_mac_address(ip=addr[0])
                    if self._on_discover:
                        self._on_discover(addr[0], mac, is_nhc2)
        server.close()
        self._on_done()



from .helpers import status_prop_in_object_is_on, extract_status_object, extract_brightness_object
from .coco_entity import CoCoEntity
import json

TOPIC_SUFFIX_CMD = '/control/devices/cmd'


class CoCoLight(CoCoEntity):

    @property
    def is_on(self):
        return self._state

    @property
    def is_on(self):
        return self._state

    @property
    def support_brightness(self):
        return self._type == 'dimmer'

    def __init__(self, dev, callback_container, client, profile_creation_id):
        super().__init__(dev, callback_container, client, profile_creation_id)
        self.update_dev(dev, callback_container)

    def turn_on(self):
        self._change_status('On')

    def turn_off(self):
        self._change_status('Off')

    def set_brightness(self, brightness):
        self._change_brightness(brightness)


    def update_dev(self, dev, callback_container=None):
        has_changed = super().update_dev(dev, callback_container)
        status_object = extract_status_object(dev)
        if self._check_for_status_change(status_object):
            self._state = status_prop_in_object_is_on(status_object)
            has_changed = True
        if self.support_brightness():
            brightness_object = extract_brightness_object(dev)
            if self._check_for_brightness_change(brightness_object):
                self._brightness = status_prop_in_object_is_on(status_object)
                has_changed = True
        return has_changed

    def _update(self, dev):
        has_changed = self.update_dev(dev)
        if has_changed:
            self._state_changed()

    def _check_for_status_change(self, property_object_with_status):
        return property_object_with_status \
               and 'Status' in property_object_with_status \
               and self._state != (status_prop_in_object_is_on(property_object_with_status))

    def _check_for_brightness_change(self, property_object_with_status):
        return property_object_with_status \
               and 'Brightness' in property_object_with_status \
               and self._brightness != (property_object_with_status.get('Brightness'))

    def _change_status(self, status):
        command = {"Method": "devices.control", "Params": [
            {"Devices": [{"Properties": [{"Status": status}], "Uuid": self._uuid}]}
        ]}
        self._client.publish(self._profile_creation_id + TOPIC_SUFFIX_CMD, json.dumps(command), 1)

    def _change_brightness(self, brightness):
        command = {"Method": "devices.control", "Params": [
            {"Devices": [{"Properties": [{"Brightness": brightness}], "Uuid": self._uuid}]}
        ]}
        self._client.publish(self._profile_creation_id + TOPIC_SUFFIX_CMD, json.dumps(command), 1)




def extract_devices(response):
    params = response['Params']
    param_with_devices = next(filter((lambda x: x and 'Devices' in x), params), None)
    return param_with_devices['Devices']


def extract_status_object(dev):
    if dev and 'Properties' in dev:
        properties = dev['Properties']
        if properties:
            return next(filter((lambda x: x and 'Status' in x), properties), None)
        else:
            return None
    else:
        return None

def extract_brightness_object(dev):
    if dev and 'Properties' in dev:
        properties = dev['Properties']
        if properties:
            return next(filter((lambda x: x and 'Brightness' in x), properties), None)
        else:
            return None
    else:
        return None


def status_prop_in_object_is_on(property_object_with_status):
    return property_object_with_status['Status'] == 'On'



import asyncio

from nhc2_coco.coco_discover_profiles import CoCoDiscoverProfiles

disc = CoCoDiscoverProfiles()

loop = asyncio.get_event_loop()


def print_u(text):
    print('\033[4m' + text +'\033[0m')


print('Searching for NiKo Home Control Controllers and profiles on them...')
try:
    results = loop.run_until_complete(disc.get_all_profiles())
finally:
    loop.run_until_complete(loop.shutdown_asyncgens())
    loop.close()

print('Done searching controllers. %d controller(s) found.' % (len(results)))
print('')
for i, result in enumerate(results):
    print_u('Controler #%d' % i)
    print(' IP: %s' % result[0])
    print(' MAC: %s' % result[1])
    profiles = result[2]
    print(' %d profile(s) found.' % (len(profiles)))
    for j, profile in enumerate(profiles):
        print(' Profile #%d:' % j)
        print('  uuid: %s' % profile.get('Uuid'))
        print('  Name: %s' % profile.get('Name'))
        print('  Type: %s' % profile.get('Type'))




import asyncio

from nhc2_coco.coco_discover import CoCoDiscover
from nhc2_coco.coco_profiles import CoCoProfiles

loop = asyncio.get_event_loop()

class CoCoDiscoverProfiles:
    """CoCoDiscover will help you discover NHC2 Profiles on all devices on the network. It will NOT find hobby
    profiles. The username then is provided by Niko (eg. hobby) This relies on not publicly documented API calls! It
    also broadcasts a UDP packet on all available (ipV4) broadcast addresses.
    """

    def __init__(self):
        self._controllers_found = []
        self._profiles_found = []
        self._done_scanning_profiles = asyncio.Event()
        CoCoDiscover(self._discover_controllers_callback, self._done_discovering_controllers_callback)

    def _done(self):
        self._done_scanning_profiles.set()

    async def _wait_until_done(self):
        await self._done_scanning_profiles.wait()

    async def get_all_profiles(self):
        await self._wait_until_done()
        return self._profiles_found

    def _discover_profiles_callback(self, address, mac):
        def inner_function(profiles):
            self._profiles_found.append((address, mac, profiles))

        return inner_function

    def _done_discovering_controllers_callback(self):
        for ctrl in self._controllers_found:
            CoCoProfiles(self._discover_profiles_callback(ctrl[0], ctrl[1]), ctrl[0],
                         self._done_discovering_profiles_callback)

    def _done_discovering_profiles_callback(self):
        if len(self._controllers_found) == len(self._profiles_found):
            loop.call_soon_threadsafe(callback=self._done)

    def _discover_controllers_callback(self, address, mac, is_nhc2):
        if (is_nhc2):
            self._controllers_found.append((address, mac))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nhc2_coco']

package_data = \
{'': ['*']}

install_requires = \
['get-mac==0.8.2', 'netifaces==0.10.9', 'paho-mqtt==1.4.0']

setup_kwargs = {
    'name': 'nhc2-coco',
    'version': '0.1.1',
    'description': 'Python controller for a Niko Home Control II installation',
    'long_description': "# pynhc2\n\nLicense: MIT\n\n## Usage\n\n### Create a NHC2 object\n\n```\nNHC2(address, username, password, port, ca_path)\n```\n\n* __address__ - IP or host of the connected controller \n* __username__ - The UUID of the profile\n* __password__ - The password\n* __port__ - (optional) The MQTT port. Default = 8883\n* __ca_path__ - (optional) Path of the CA file. Default = included CA file.\n\n example:\n\n ```\n coco = NHC2('192.168.1.2', 'abcdefgh-ijkl-mnop-qrst-uvwxyz012345', 'secret_password')\n ```\n \n### What now?\n TODO - write doc....\n \n## What can you do to help?\n\n * Contribute to this project with constructive issues, suggestions, PRs, etc.\n * Help me in any way to get support for more entities (eg dimmer)\n",
    'author': 'filipvh',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/filipvh/nhc2-coco',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



import paho.mqtt.client as mqtt

MQTT_TLS_VERSION = 2
MQTT_PROTOCOL = mqtt.MQTTv311
MQTT_TRANSPORT = "tcp"

MQTT_RC_CODES = ['',
                 'Connection refused - incorrect protocol version',
                 'Connection refused - invalid client identifier',
                 'Connection refused - server unavailable',
                 'Connection refused - bad username or password',
                 'Connection refused - not authorised']

MQTT_TOPIC_SUFFIX_CMD = '/control/devices/cmd'
MQTT_TOPIC_SUFFIX_RSP = '/control/devices/rsp'
MQTT_TOPIC_SUFFIX_EVT = '/control/devices/evt'
MQTT_TOPIC_SUFFIX_SYS_EVT = '/system/evt'
MQTT_TOPIC_PUBLIC_CMD = 'public/system/cmd'
MQTT_TOPIC_PUBLIC_RSP = 'public/system/rsp'
MQTT_TOPIC_PUBLIC_AUTH_CMD = 'public/authentication/cmd'
MQTT_TOPIC_PUBLIC_AUTH_RSP = 'public/authentication/rsp'



import json
import os
from time import sleep

import paho.mqtt.client as mqtt

from nhc2_coco.const import MQTT_TOPIC_PUBLIC_AUTH_RSP, MQTT_PROTOCOL, MQTT_TRANSPORT, MQTT_TOPIC_PUBLIC_AUTH_CMD


class CoCoProfiles:
    """CoCoDiscover will collect a list of profiles on a NHC2
    """

    def __init__(self, callback, address, done_discovering_profiles_callback, port=8883, ca_path=None):

        if ca_path is None:
            ca_path = os.path.dirname(os.path.realpath(__file__)) + '/coco_ca.pem'
        client = mqtt.Client(protocol=MQTT_PROTOCOL, transport=MQTT_TRANSPORT)
        client.tls_set(ca_path)
        client.tls_insecure_set(True)
        self._client = client
        self._address = address
        self._callback = callback
        self._done_discovering_profiles_callback = done_discovering_profiles_callback
        self._loop = 0
        self._port = port
        self._client.on_message = self._on_message
        self._client.on_connect = self._on_connect
        self._client.loop_start()
        self._client.connect_async(self._address, self._port)
        while self._loop < 50 and self._loop >= 0:
            sleep(0.5)
            self._client.loop(max_packets=10)
        self._client.disconnect()
        self._client.loop_stop()
        if (self._loop > 0):
            self._callback([])
            self._loop = -1
        self._done_discovering_profiles_callback()

    def _on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            client.subscribe(MQTT_TOPIC_PUBLIC_AUTH_RSP, qos=1)
            client.publish(MQTT_TOPIC_PUBLIC_AUTH_CMD, '{"Method":"profiles.list"}', 1)
        else:
            self._callback([])
            self._loop = -1

    def _on_message(self, client, userdata, message):

        topic = message.topic
        response = json.loads(message.payload)
        if topic == MQTT_TOPIC_PUBLIC_AUTH_RSP \
                and response.get('Method') == 'profiles.list' \
                and 'Params' in response \
                and (len(response.get('Params')) == 1) \
                and 'Profiles' in response.get('Params')[0]:
            self._loop = -1
            self._callback(response.get('Params')[0].get('Profiles'))
